open System
open Future 
open Mpi


module type S = sig
  type 'a t
  val tabulate : (int -> 'a) -> int -> 'a t
  val seq_of_array : 'a array -> 'a t
  val array_of_seq : 'a t -> 'a array
  val iter: ('a -> unit) -> 'a t -> unit
  val length : 'a t -> int
  val empty : unit  ->'a t
  val cons : 'a -> 'a t -> 'a t
  val singleton : 'a -> 'a t
  val append : 'a t -> 'a t -> 'a t
  val nth : 'a t -> int -> 'a
  val map : ('a -> 'b) -> 'a t -> 'b t
  val map_reduce : ('a -> 'b) -> ('b -> 'b -> 'b) -> 'b -> 'a t -> 'b
  val reduce : ('a -> 'a -> 'a) -> 'a -> 'a t -> 'a
  val flatten : 'a t t -> 'a t
  val repeat : 'a -> int -> 'a t
  val zip : ('a t * 'b t) -> ('a * 'b) t
  val split : 'a t -> int -> 'a t * 'a t
  val scan: ('a -> 'a -> 'a) -> 'a -> 'a t -> 'a t
end



(*******************************************************)
(* Sequential Sequences Based on a List Representation *)
(*******************************************************)

module ListSeq : S = struct

  type 'a t = 'a list

  let length = List.length

  let empty () = []

  let cons (x:'a) (s:'a t) = x::s

  let singleton x = [x]

  let append = List.append

  let tabulate f n =
    let rec helper acc x =
      if x = n then List.rev acc
      else helper ((f x)::acc) (x+1) in
    helper [] 0

  let nth = List.nth

  let filter = List.filter

  let map = List.map

  let reduce = List.fold_left

  let map_reduce m r b s = reduce r b (map m s)

  let repeat x n =
    let rec helper x n acc =
      if n = 0 then acc else helper x (n-1) (x::acc) in
    helper x n []

  let flatten = List.flatten

  let zip (s1,s2) = List.combine s1 s2

  let split s i =
    let rec helper s i acc =
      match s,i with
        | [],_ -> failwith "split"
        | _,0 -> (List.rev acc,s)
        | h::t,_ -> helper t (i-1) (h::acc) in
    helper s i []

  let iter = List.iter

  let array_of_seq = Array.of_list

  let seq_of_array = Array.to_list

  let scan f b s = 
    let (_,xs) = List.fold_left (fun (v,ls) e -> let r = f v e in (r,r::ls)) (b,[]) s in
    List.rev xs

end


(*******************************************************)
(* Parallel Sequences                                  *)
(*******************************************************)

module type SEQ_ARGS = sig 
  val use_mpi: bool
end


module Seq (Par : Future.S) (Arg : SEQ_ARGS) : S = struct

  type 'a t = 'a array option  (* a possible representation of sequences -- feel free to change *)

  (* useful helper code -- feel free to change *)

  (* number of cores on your machine *)
  let num_cores = System.cpu_count ()

  (* Compute f using (low,high) indices for chunks of numbers from 1 to n *)
  (* In general, f might create a future.  That future might perform a computation
     across elements of an array from lo to hi-1. *)
  let chunks (f: int * int -> 'a) (n:int): 'a array = 
    let chunks = min (num_cores) n in 
    Array.init chunks (fun i -> 
      let lo = (n * i) / chunks in 
      let hi = (n * (i+1) / chunks) - 1 in 
      f (lo,hi) 
    )

  (* Copy an array of arrays into a single array using Array.blit *)
  let fast_append (arrs: ('a array) array) (len: int) : 'a array = 
    let curr = ref arrs.(0) in 
    let clen = Array.length arrs in 
    if clen = 1 then !curr
    else 
      let default = (!curr).(0) in 
      let result = Array.make len default in
      let idx = ref 0 in 
      for i = 0 to clen - 1 do 
        (curr := if i = 0 then !curr else arrs.(i);
        let l = Array.length !curr in 
        Array.blit !curr 0 result !idx l;
        idx := !idx + l)
      done; result

  let tabulate f n = 
    if n <= 0 then None 
    else
      let chunks = chunks (fun (lo,hi) -> Par.future (Array.init (hi-lo+1)) (fun i -> f (i+lo))) n in
      Some (fast_append (Array.map Par.force chunks) n)


  let seq_of_array a =
    if (Array.length a) = 0 then None else Some a


  let array_of_seq seq = 
    match seq with 
      | None -> [| |]
      | Some arr -> arr


  let iter f seq = failwith "implement me"


  let length seq = 
    match seq with
    | None -> 0
    | Some a -> Array.length a


  let empty () = None


  let cons elem seq = failwith "implement me"


  let singleton elem = Some (Array.make 1 elem)


  let append seq1 seq2 = failwith "implement me"


  let nth seq i = failwith "implement me"


  let map f seq = failwith "implement me"


  let map_reduce m r b seq = failwith "implement me"


  let reduce r = failwith "implement me"


  let flatten seqseq = failwith "implement me"


  let repeat elem num = failwith "implement me"


  let zip (seq1,seq2) = failwith "implement me"


  let split seq x = failwith "implement me"


  (*******************************************************)
  (* Parallel Prefix Sum                                 *)
  (*******************************************************)

  (* Here you will implement a version of the parallel prefix scan for a sequence 
   * [a0; a1; ...], the result of scan will be [f base a0; f (f base a0) a1; ...] *)
  let scan (f: 'a -> 'a -> 'a) (base: 'a) (seq: 'a t) : 'a t =
    failwith "implement me"
        
end







